
/**************************************************************************/
/*                                                                        */
/* This subroutin computes next values in subdomain of process me         */ 
/* within the domain                                                      */
/*                                                                        */
/**************************************************************************/

void Explicit(double** x0, double** x, double dt, double* r, double hx, double hy, int me, int* xs, int* ys, int* xe, int* ye, double k0)
{ 
  int i, j;
  double diagx, diagy, weightx, weighty, rk;
  /* 
     The stencil of the explicit operator for the heat equation
     on a regular rectangular grid using a five point finite difference
     scheme in space is

     |         + 1                 |
     |                             |
     dt/(h*h) *  | +1      -4 + h*h/dt     + 1 |
     |                             |
     |          + 1                |
   */

  diagx = - 2.0 + hx*hx/(2*k0*dt);
  diagy = - 2.0 + hy*hy/(2*k0*dt);
  weightx = k0 * dt/(hx*hx);
  weighty = k0 * dt/(hy*hy);

  /* perform an explicit update on the points within the domain */

  for(i=ys[me];i<=ye[me];i++)
  { for(j=xs[me];j<=xe[me];j++)
    {
      x[i][j] = weighty *( x0[i-1][j] + x0[i+1][j] + x0[i][j]*diagy) + weightx *( x0[i][j-1] + x0[i][j+1] + x0[i][j]*diagx);
    }
  }

  /* copy back the computed value : x  <-- x^(n+1)             */
  /*                                x0 <-- x^n                 */
  /* and compute at the same time the 2_norm of the "residual" */

  *r = 0.0;

  for(i=ys[me];i<=ye[me];i++)
  { for(j=xs[me];j<=xe[me];j++)
    {
      rk = x0[i][j] - x[i][j];
      *r  = *r + rk * rk;
      x0[i][j] = x[i][j];
    }
  }
}

/**************************************************************************/
/*                                                                        */
/* This subroutine setups the initial guess, i.e. the initial temperature */
/* within the domain                                                      */
/*                                                                        */
/**************************************************************************/

void initvalues(double** x0, int size_tot_x, int size_tot_y, double temp1_init, double temp2_init)
{      
  int i, j;

  for(j=0;j<=size_tot_x-1;j++)
  { 
    x0[0][j] = temp1_init;
    x0[size_tot_y-1][j] = temp1_init;
  }

  for(j=0;j<=size_tot_y-1;j++)
  { 
    x0[j][0] = temp1_init;
    x0[j][size_tot_x-1] = temp1_init;
  }

  for(j=1;j<=size_tot_x-2;j++)
  { 
    x0[1][j] = temp1_init;
    x0[size_tot_y-2][j] = temp1_init;
  }

  for(j=0;j<=size_tot_y-2;j++)
  { 
    x0[j][1] = temp1_init;
    x0[j][size_tot_x-2] = temp1_init;
  }

  for(j=2;j<=size_tot_y-3;j++)
  { 
    for(i=2;i<=size_tot_x-3;i++)
    {
      x0[j][i] = temp2_init;
    }
  }    
}
