#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <CL/cl.h>
#include "explicitGPU.h"

#define min(a,b) a <= b ? a : b

int main(void) {

  // kernel code

  const char *source ="__kernel void line_compute(const double diagx, const double diagy,\
                       const double weightx,  const double weighty,  const int size_x,\
                       __global double* tab_current,  __global double* tab_new, __global double* r)\
                       { int iy = get_global_id(0)+1;\
                         int ix = get_global_id(1)+1;\
                           double new_value, cell, cell_n, cell_s, cell_w, cell_e;\
                           double rk;\
                           int i,j;\
                           cell_s = tab_current[(iy+1)*(size_x+2)+ix];\
                           cell_n = tab_current[(iy-1)*(size_x+2)+ix];\
                           cell_e = tab_current[iy*(size_x+2)+(ix+1)];\
                           cell_w = tab_current[iy*(size_x+2)+(ix-1)];\
                           cell     = tab_current[iy*(size_x+2)+ix];\
                           new_value = weighty *( cell_n + cell_s + cell*diagy)+\
                           weightx *( cell_e + cell_w + cell*diagx);\
                           rk = cell - new_value;\
                           r[iy*(size_x+2)+ix] = rk *rk;\
                           tab_new[iy*(size_x+2)+ix] = new_value;\
                       }";

  /* size of the discretization */

  int size_x, size_y, size_tot_x, size_tot_y, maxStep;
  double error;
  double* x_input;
  double* x_output;
  double* r;      
  size_t source_size = strlen(source);

  double dt1, dt2, dt, hx, hy, result, epsilon, k0;
  double diagx, diagy, weightx, weighty;

  /* File descriptor */
  FILE* file1;

  /* convergence */
  int convergence = 0;

  /* index loop */

  int i, j, step;
  double t;

  /* Variables for Clock */

  time_t time_init, time_final;
  double elapsed_time;

  /* temp1_init: temperature init on edges */

  double temp1_init = 10.0;

  /* temp2_init: temperature init inside */

  double temp2_init = -10.0;

  /* diffusivity coefficient */

  k0 = 1;

  /* kernel source */

  printf("Size x of the square \n");
  scanf("%d",&size_x);
  printf("Size y of the square \n");
  scanf("%d",&size_y);
  printf("Max. number of steps \n");
  scanf("%d",&maxStep);
  printf("Time step\n");
  scanf("%lf",&dt1);
  printf("Convergence \n");
  scanf("%lf",&epsilon);

  hx  = 1.0/(double)(size_x+2);
  hy  = 1.0/(double)(size_y+2);
  dt2 = 0.25*((min(hx,hy))*(min(hx,hy)));

  if(dt1>=dt2)
  { printf("\n");
    printf("Time step too large in param file - Taking convergence criterion\n");
    dt=dt2; }
  else { dt=dt1; }

  size_tot_x = size_x+2;
  size_tot_y = size_y+2;

  // Main values array
  x_input  = malloc((size_tot_x)*(size_tot_y)*sizeof(double));
  x_output = malloc((size_tot_x)*(size_tot_y)*sizeof(double));

  // Allocate cumulative error array
  r = malloc((size_tot_x)*(size_tot_y)*sizeof(double));

  /* Array initialization */

  initvalues(x_input, size_x, size_y, temp1_init, temp2_init );
  initvalues(x_output, size_x, size_y, temp1_init, temp2_init );

  diagx = - 2.0 + hx*hx/(2*k0*dt);
  diagy = - 2.0 + hy*hy/(2*k0*dt);
  weightx = k0 * dt/(hx*hx);
  weighty = k0 * dt/(hy*hy);

  step = 0;
  t    = 0.0;

  //time_init=clock();
	time_init = time(NULL);

  /* OpenCL initialization */

  // Get platform and device information
  cl_platform_id platform_id = NULL;
  cl_device_id device_id = NULL;   
  cl_uint ret_num_devices;
  cl_uint ret_num_platforms;
  cl_int ret = clGetPlatformIDs(1, &platform_id, &ret_num_platforms);
  ret = clGetDeviceIDs( platform_id, CL_DEVICE_TYPE_GPU, 1, 
      &device_id, &ret_num_devices);

  // Create an OpenCL context
  cl_context context = clCreateContext( NULL, 1, &device_id, NULL, NULL, &ret);

  // Create a command queue
  cl_command_queue command_queue = clCreateCommandQueue(context, device_id, 0, &ret);

  // Create memory buffers on the device for each vector 
  cl_mem x0_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY,
      (size_x+2) * (size_y+2) * sizeof(cl_double), NULL, &ret);	    
  cl_mem x_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY,
      (size_x+2) * (size_y+2) * sizeof(cl_double), NULL, &ret);	    
  cl_mem r_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY,
      (size_x+2) * (size_y+2) * sizeof(cl_double), NULL, &ret);	    

  // Write to buffers
  ret = clEnqueueWriteBuffer(command_queue, x0_mem_obj, CL_TRUE, 0, 
      (size_x+2) * (size_y+2) * sizeof(cl_double), x_input, 0, NULL, NULL);	   	    
  ret = clEnqueueWriteBuffer(command_queue, x_mem_obj, CL_TRUE, 0, 
      (size_x+2) * (size_y+2) * sizeof(cl_double), x_output, 0, NULL, NULL);	   	    
  ret = clEnqueueWriteBuffer(command_queue, r_mem_obj, CL_TRUE, 0, 
      (size_x+2) * (size_y+2) * sizeof(cl_double), r, 0, NULL, NULL);	   	    

  // Create a program from the kernel source
  cl_program program = clCreateProgramWithSource(context, 1, 
      (const char **)&source, (const size_t *)&source_size, &ret);

  // Build the program
  if (clBuildProgram(program, 1, &device_id, NULL, NULL, NULL) != CL_SUCCESS)
  { char buffer[1024];
    clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, NULL);
    fprintf(stderr, "CL Compilation failed:\n%s", buffer);
    abort(); }

    // Create the OpenCL kernel
    cl_kernel kernel = clCreateKernel(program, "line_compute", &ret);

    // Set the arguments of the kernel
    ret = clSetKernelArg(kernel, 0, sizeof(double), (void *)&diagx);
    ret = clSetKernelArg(kernel, 1, sizeof(double), (void *)&diagy);
    ret = clSetKernelArg(kernel, 2, sizeof(double), (void *)&weightx);
    ret = clSetKernelArg(kernel, 3, sizeof(double), (void *)&weighty);
    ret = clSetKernelArg(kernel, 4, sizeof(int), (void *)&size_x);
    ret = clSetKernelArg(kernel, 5, sizeof(cl_mem), (void *)&x0_mem_obj);    
    ret = clSetKernelArg(kernel, 6, sizeof(cl_mem), (void *)&x_mem_obj);    
    ret = clSetKernelArg(kernel, 7, sizeof(cl_mem), (void *)&r_mem_obj);

    // Execute the OpenCL kernel on the list
    size_t global_item_size[2] = {size_y, size_x}; 
    size_t local_item_size[2] = {8, 8}; 

    // Main loop

    while(!convergence)
    {  
      step = step + 1;
      t    = t + dt;

      Explicit(command_queue, global_item_size, local_item_size, kernel, x0_mem_obj, x_mem_obj, r_mem_obj, x_input, r, size_x, size_y, &error) ; 

      result = sqrt(error);

      if ((result<epsilon) || (step>maxStep)) break;
    }

    time_final = time(NULL);
    elapsed_time = (time_final-time_init);

    printf("\n");
    printf("  Time step = %3.18f\n", dt);
    printf("\n");
    printf("  Convergence = %11.9f after %d steps\n", epsilon, step);
    printf("\n");
    printf("  Problem size = %d\n", size_x*size_y);
    printf("\n");
    printf("  Wall Clock = %15.6f\n", elapsed_time);

    /* Print the solution at each point of the grid */

    printf("\n");
    printf("  Computed solution in outputGPU.dat\n");
    printf("\n");

    file1=fopen("outputGPU.dat","w");

    for(j=size_y+1;j>=0;j--)
    { 
      for (i=0;i<=size_x;i++)
      { fprintf(file1,"%15.11f ",x_input[j*(size_x+2)+i]);}
      fprintf(file1,"%15.11f", x_input[j*(size_x+2)+size_x+1]);
      fprintf(file1,"\n");
    }
    fclose(file1);

    free(x_input);
    free(r);

    // OpenCL Clean up
    ret = clFinish(command_queue);
    ret = clReleaseKernel(kernel);
    ret = clReleaseProgram(program);
    ret = clReleaseMemObject(x0_mem_obj);    
    ret = clReleaseMemObject(x_mem_obj);    
    ret = clReleaseMemObject(r_mem_obj);
    ret = clReleaseCommandQueue(command_queue);
    ret = clReleaseContext(context);

    return 0;

}
